const std = @import("std");
const builtin = @import("builtin");
const math = @import("math.zig");
const graphics = @import("graphics.zig");
const player = @import("player.zig");

const Color = graphics.Color;
const Canvas = graphics.Canvas(640, 480);
const Player = player.Player;

const Vec = math.Vec;
const Vec2 = Vec(2, f32);

const Ray = math.Ray;
const Ray2 = Ray(2, f32);

const Mat = math.Mat;

pub const Game = struct {
    canvas: Canvas,
    player: Player,
    camera: Vec2,
    map: Mat(24, 24, u8),
    frame: u32,

    pub fn init() Game {
        return .{
            .canvas = Canvas.init(),
            .player = Player.init(Vec2.init(12, 12), Vec2.init(-1, 0)),
            .camera = Vec2.init(0, 0.66),
            .frame = 0,
            .map = Mat(24, 24, u8).init(.{
                .{ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
                .{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 3, 3, 3, 3, 3, 0, 0, 0, 1 },
                .{ 1, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 1 },
                .{ 1, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 3, 0, 0, 0, 3, 0, 0, 0, 1 },
                .{ 1, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 1 },
                .{ 1, 0, 0, 0, 0, 0, 2, 2, 0, 2, 2, 0, 0, 0, 0, 3, 3, 3, 3, 3, 0, 0, 0, 1 },
                .{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 4, 0, 4, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 4, 0, 0, 0, 0, 5, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 4, 0, 4, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 4, 0, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
                .{ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
            }),
        };
    }

    pub fn update(self: *Game, dt: f32) void {
        self.frame += 1;
        self.player.update(dt);
        self.camera = Vec2.init(0, 0.66).rotate(self.player.direction.angle() - std.math.pi);
    }

    pub fn draw(self: *Game) void {
        for (0..self.canvas.width()) |x| {
            const cameraX: f32 = 2.0 * @as(f32, @floatFromInt(x)) / @as(f32, @floatFromInt(self.canvas.width())) - 1.0;
            const rayDir = self.player.direction.add(self.camera.mulScalar(cameraX));
            var mapPos = self.player.position.to(usize);
            const deltaDist = Vec2.init(
                if (rayDir.x() == 0.0) std.math.floatMax(f32) else @abs(1.0 / rayDir.x()),
                if (rayDir.y() == 0.0) std.math.floatMax(f32) else @abs(1.0 / rayDir.y()),
            );
            const step = Vec(2, isize).init(
                if (rayDir.x() < 0.0) -1 else 1,
                if (rayDir.y() < 0.0) -1 else 1,
            );
            var sideDist = Vec2.init(
                if (rayDir.x() < 0.0)
                    (self.player.position.x() - @as(f32, @floatFromInt(mapPos.x()))) * deltaDist.x()
                else
                    (@as(f32, @floatFromInt(mapPos.x())) + 1.0 - self.player.position.x()) * deltaDist.x(),
                if (rayDir.y() < 0.0)
                    (self.player.position.y() - @as(f32, @floatFromInt(mapPos.y()))) * deltaDist.y()
                else
                    (@as(f32, @floatFromInt(mapPos.y())) + 1.0 - self.player.position.y()) * deltaDist.y(),
            );

            var hit = false;
            var side: u32 = 0;
            while (!hit) {
                if (sideDist.x() < sideDist.y()) {
                    sideDist.data[0] += deltaDist.x();
                    mapPos.data[0] = if (step.x() < 0) mapPos.x() - 1 else mapPos.x() + 1;
                    side = 0;
                } else {
                    sideDist.data[1] += deltaDist.y();
                    mapPos.data[1] = if (step.y() < 0) mapPos.y() - 1 else mapPos.y() + 1;
                    side = 1;
                }
                // WHY?
                if (mapPos.x() < 0 or mapPos.x() >= self.map.Cols or mapPos.y() < 0 or mapPos.y() >= self.map.Rows) {
                    break; // Out of bounds
                }
                if (self.map.get(mapPos.x(), mapPos.y()) > 0) {
                    hit = true;
                }
            }

            if (hit) {
                const perpWallDist = if (side == 0)
                    (@as(f32, @floatFromInt(mapPos.x())) - self.player.position.x() + (@as(f32, @floatFromInt(1 - step.x()))) / 2) / rayDir.x()
                else
                    (@as(f32, @floatFromInt(mapPos.y())) - self.player.position.y() + (@as(f32, @floatFromInt(1 - step.y()))) / 2) / rayDir.y();

                const h: f32 = @floatFromInt(self.canvas.height());

                const lineHeight = h / perpWallDist;
                var start = -lineHeight / 2 + h / 2;
                if (start < 0) start = 0;
                var end = lineHeight / 2 + h / 2;
                if (end >= h) end = h - 1;
                self.canvas.drawLine(Vec2.init(@floatFromInt(x), (0)), Vec2.init(@floatFromInt(x), (start)), Color.BG);
                self.canvas.drawLine(Vec2.init(@floatFromInt(x), (start)), Vec2.init(@floatFromInt(x), (end)), Color.shade(Color.FG2, lineHeight));
                self.canvas.drawLine(Vec2.init(@floatFromInt(x), (end)), Vec2.init(@floatFromInt(x), (h)), Color.FG);
            }
        }
    }
};
test "Game update" {
    var g = Game.init();
    g.update(1.0);
    try std.testing.expectEqual(g.frame, 1);
}
test "Game draw" {
    var g = Game.init();
    g.draw();
    try std.testing.expectEqual(g.frame, 0);
}
