const std = @import("std");
const math = @import("math.zig");

const Vec = math.Vec;
const Vec2 = Vec(2, f32);
const Vec3 = Vec(3, f32);
const Vec4 = Vec(4, f32);

pub const Color = packed struct {
    red: u8,
    green: u8,
    blue: u8,
    alpha: u8,
    pub const FG: Color = Color{ .red = 32, .green = 56, .blue = 15, .alpha = 255 };
    pub const FG2: Color = Color{ .red = 48, .green = 98, .blue = 48, .alpha = 128 };
    pub const FG3: Color = Color{ .red = 48, .green = 98, .blue = 48, .alpha = 255 };
    pub const BG: Color = Color{ .red = 139, .green = 172, .blue = 15, .alpha = 128 };
    pub const RED: Color = Color{ .red = 255, .green = 0, .blue = 0, .alpha = 255 };
    pub const WHITE: Color = Color{ .red = 255, .green = 255, .blue = 255, .alpha = 255 };
    pub const BLACK: Color = Color{ .red = 0, .green = 0, .blue = 0, .alpha = 255 };
    pub const BLUE: Color = Color{ .red = 0, .green = 0, .blue = 255, .alpha = 255 };
    pub const PINK: Color = Color{ .red = 255, .green = 192, .blue = 203, .alpha = 255 };
    pub const PURPLE: Color = Color{ .red = 128, .green = 0, .blue = 130, .alpha = 255 };
    pub const GREEN: Color = Color{ .red = 0, .green = 255, .blue = 0, .alpha = 255 };
    pub const WALL: Color = Color{ .red = 110, .green = 115 + 30, .blue = 81, .alpha = 255 };
    pub const CEILING: Color = Color{ .red = 76, .green = 185, .blue = 124, .alpha = 255 };
    pub const FLOOR: Color = Color{ .red = 59, .green = 61 + 33, .blue = 43, .alpha = 255 };
    pub fn eq(a: Color, b: Color) bool {
        return a.red == b.red and a.green == b.green and a.blue == b.blue and a.alpha == b.alpha;
    }
    pub fn shade(a: Color, distance: f32) Color {
        const maxShadeDistance: f32 = 250.0; // Adjust this value based on your game's scale

        const shadingFactor: f32 = std.math.clamp(1.0 - (distance / maxShadeDistance), 0.0, 1.0);

        const fr: f32 = @floatFromInt(a.red);
        const rr: u8 = @intFromFloat(fr * shadingFactor);

        const fg: f32 = @floatFromInt(a.green);
        const rg: u8 = @intFromFloat(fg * shadingFactor);

        const fb: f32 = @floatFromInt(a.blue);
        const rb: u8 = @intFromFloat(fb * shadingFactor);

        return Color{
            .red = rr,
            .green = rg,
            .blue = rb,
            .alpha = a.alpha,
        };
    }
};

pub fn Canvas(comptime _width: comptime_int, comptime _height: comptime_int) type {
    return struct {
        const Self = @This();

        buffer: [_width * _height]Color = undefined,

        pub fn init() Self {
            return .{};
        }
        pub fn clear(self: *Self, color: Color) void {
            @memset(&self.buffer, color);
        }
        pub fn print(self: *Self) void {
            for (0.._height - 1) |y| {
                for (0.._width - 1) |x| {
                    var color = self.buffer[y * _width + x];
                    if (color.eq(Color.FG)) {
                        std.debug.print(" ", .{});
                    } else if (color.eq(Color.BG)) {
                        std.debug.print("█", .{});
                    } else if (color.eq(Color.BG2)) {
                        std.debug.print("▒", .{});
                    } else {
                        std.debug.print("█", .{});
                    }
                }
                std.debug.print("\n", .{});
            }
        }
        pub fn width(self: *const Self) usize {
            _ = self;
            return _width;
        }
        pub fn height(self: *const Self) usize {
            _ = self;
            return _height;
        }
        pub fn drawPixel(self: *Self, x: usize, y: usize, color: Color) void {
            if (x >= _width or y >= _height) return;
            if (x < 0 or y < 0) return;
            self.buffer[y * _width + x] = color;
        }
        pub fn drawLine(self: *Self, start: Vec2, end: Vec2, color: Color) void {
            if (start.eq(end)) return;

            const delta = end.sub(start);

            const xDir = Vec2.init(if (start.x() < end.x()) 1 else -1, 0);
            const yDir = Vec2.init(0, if (start.y() < end.y()) 1 else -1);

            var err = @abs(delta.x()) - @abs(delta.y());
            const decision = 2 * err;

            var pen = start;
            while (true) {
                if (end.eq(pen)) break;
                if (pen.x() < 0 or pen.y() < 0) break;
                if (pen.x() >= _width or pen.y() >= _height) break;

                self.drawPixel(@intFromFloat(pen.x()), @intFromFloat(pen.y()), color);

                if (decision > -delta.y()) {
                    err -= delta.y();
                    pen = pen.add(xDir);
                }
                if (decision < delta.x()) {
                    err += delta.x();
                    pen = pen.add(yDir);
                }
            }
        }
        pub fn drawRect(self: *Self, pos: Vec2, w: f32, h: f32, color: Color) void {
            self.drawLine(pos, pos.add(Vec2.init(w, 0)), color);
            self.drawLine(pos, pos.add(Vec2.init(0, h)), color);
            self.drawLine(pos.add(Vec2.init(w, 0)), pos.add(Vec2.init(w, h)), color);
            self.drawLine(pos.add(Vec2.init(0, h)), pos.add(Vec2.init(w, h)), color);
        }
        pub fn fillRect(self: *Self, pos: Vec2, w: f32, h: f32, color: Color) void {
            const fromX: u32 = @intFromFloat(pos.x());
            const fromY: u32 = @intFromFloat(pos.y());
            const toX = fromX + @as(u32, @intFromFloat(w));
            const toY = fromY + @as(u32, @intFromFloat(h));
            for (fromX..toX) |xx| {
                for (fromY..toY) |yy| {
                    self.drawPixel(@intCast(xx), @intCast(yy), color);
                }
            }
        }
    };
}
test "Canvas init" {
    const canvas = Canvas(10, 10).init();
    try std.testing.expectEqual(canvas.buffer.len, 100);
}
test "Canvas clear" {
    var canvas = Canvas(10, 10).init();
    canvas.clear(Color.FG);
    try std.testing.expectEqual(std.meta.eql(canvas.buffer, [_]Color{Color.FG} ** 100), true);
}
test "Canvas width" {
    var canvas = Canvas(10, 10).init();
    try std.testing.expectEqual(canvas.width(), 10);
}
test "Canvas height" {
    var canvas = Canvas(10, 10).init();
    try std.testing.expectEqual(canvas.height(), 10);
}
test "Canvas drawPixel" {
    var canvas = Canvas(10, 10).init();
    canvas.drawPixel(0, 0, Color.FG);
    try std.testing.expectEqual(canvas.buffer[0], Color.FG);
}
test "Canvas drawLine" {
    var canvas = Canvas(10, 10).init();
    canvas.drawLine(Vec2.init(0, 0), Vec2.init(10, 10), Color.FG);
    try std.testing.expectEqual(canvas.buffer[0], Color.FG);
}
test "Canvas drawRect" {
    var canvas = Canvas(10, 10).init();
    canvas.drawRect(Vec2.init(0, 0), 10, 10, Color.FG);
    try std.testing.expectEqual(canvas.buffer[0], Color.FG);
    try std.testing.expectEqual(canvas.buffer[9], Color.FG);
}
test "Canvas fillRect" {
    var canvas = Canvas(10, 10).init();
    canvas.fillRect(Vec2.init(0, 0), 10, 10, Color.FG);
    try std.testing.expectEqual(canvas.buffer[0], Color.FG);
    try std.testing.expectEqual(canvas.buffer[9], Color.FG);
}
